QueryStatusFinished = "finished"
QueryStatusPending = "pending"
QueryStatusError = "error"


class QueryStatus(object):
    Finished: str = QueryStatusFinished
    Pending: str = QueryStatusPending
    Error: str = QueryStatusError



from typing import Union

from .base_error import BaseError


class QueryRunRateLimitError(BaseError):
    """
    Base class for all QueryRunRateLimitError errors.
    """

    def __init__(self):
        self.message = "QUERY_RUN_RATE_LIMIT_ERROR: you have exceeded the rate limit for creating/running new queries"
        super().__init__(self.message)


class QueryRunTimeoutError(BaseError):
    """
    Base class for all QueryRunTimeoutError errors.
    """

    def __init__(self, timeoutMinutes: Union[int, float]):
        self.message = f"QUERY_RUN_TIMEOUT_ERROR: your query has timed out after {timeoutMinutes} minutes."
        super().__init__(self.message)


class QueryRunExecutionError(BaseError):
    """
    Base class for all QueryRunExecutionError errors.
    """

    def __init__(self):
        self.message = "QUERY_RUN_EXECUTION_ERROR: an error has occured while executing your query."
        super().__init__(self.message)



from shroomdk.api import API
from shroomdk.integrations import QueryIntegration
from shroomdk.models import Query

API_BASE_URL = "https://api.flipsidecrypto.com"


class ShroomDK(object):
    def __init__(self, api_key: str, api_base_url: str = API_BASE_URL):
        self.api = API(api_base_url, api_key)

    def query(
        self,
        sql,
        ttl_minutes=60,
        cached=True,
        timeout_minutes=20,
        retry_interval_seconds=1,
        page_size=100000,
        page_number=1,
    ):
        query_integration = QueryIntegration(self.api)

        return query_integration.run(
            Query(
                sql=sql,
                ttl_minutes=ttl_minutes,
                timeout_minutes=timeout_minutes,
                retry_interval_seconds=retry_interval_seconds,
                page_size=page_size,
                page_number=page_number,
                cached=cached,
            )
        )



from datetime import datetime

from shroomdk.integrations.query_integration.query_result_set_builder import (
    QueryResultSetBuilder,
)
from shroomdk.models.api import QueryResultJson
from shroomdk.models.query_status import QueryStatus


def getQueryResultSetData(status: str) -> QueryResultJson:
    return QueryResultJson(
        queryId="test",
        status=status,
        results=[
            [1, "0x-tx-id-0", "0xfrom-address-0", True, 0.5],
            [2, "0x-tx-id-1", "0xfrom-address-1", False, 0.75],
            [3, "0x-tx-id-2", "0xfrom-address-2", False, 1.75],
            [4, "0x-tx-id-3", "0xfrom-address-3", True, 100.001],
        ],
        startedAt="2022-05-19T00:00:00.00Z",
        endedAt="2022-05-19T00:01:30.00Z",
        columnLabels=[
            "block_id",
            "tx_id",
            "from_address",
            "succeeded",
            "amount",
        ],
        columnTypes=["number", "string", "string", "boolean", "number"],
        message="",
        errors=None,
        pageSize=100,
        pageNumber=0,
    )


def test_run_stats():
    qr = QueryResultSetBuilder(getQueryResultSetData(QueryStatus.Finished))

    # Start/end are datetime objects?
    assert type(qr.run_stats.started_at) == datetime
    assert type(qr.run_stats.ended_at) == datetime

    # Elapsed seconds
    assert qr.run_stats.elapsed_seconds == 90

    # Record count
    assert qr.run_stats.record_count == 4


def test_records():
    qr = QueryResultSetBuilder(getQueryResultSetData(QueryStatus.Finished))

    # Records Length Matches Row Length?
    assert qr.records is not None
    assert qr.rows is not None
    assert qr.columns is not None
    assert len(qr.records) == len(qr.rows)

    # Column Length Matches Records Key Length
    for record in qr.records:
        assert record is not None

        assert len(record.keys()) == len(qr.columns)

    # Columns = Record Keys
    for record in qr.records:
        for column in qr.columns:
            assert column in record.keys()

    # Record values match row values?
    for record, row in zip(qr.records, qr.rows):
        for column, value in zip(qr.columns, row):
            assert record[column] == value


def test_status():
    # Status is finished?
    qr = QueryResultSetBuilder(getQueryResultSetData(QueryStatus.Finished))
    assert qr.status == QueryStatus.Finished

    # Status is pending?
    qr = QueryResultSetBuilder(getQueryResultSetData(QueryStatus.Pending))
    assert qr.status == QueryStatus.Pending

    # Status is error?
    qr = QueryResultSetBuilder(getQueryResultSetData(QueryStatus.Error))
    assert qr.status == QueryStatus.Error


def test_query_id():
    # Query ID is set?
    qr = QueryResultSetBuilder(getQueryResultSetData(QueryStatus.Finished))
    assert qr.query_id is not None

    # Query ID is test
    qr = QueryResultSetBuilder(getQueryResultSetData(QueryStatus.Finished))
    assert qr.query_id == "test"



import json

from shroomdk.api import API
from shroomdk.models import Query, QueryStatus
from shroomdk.models.api import QueryResultJson


def test_create_query_success(requests_mock):
    api = API("https://api.flipsidecrypto.xyz", "api_key")

    result = requests_mock.post(
        api.get_url("queries"),
        text=json.dumps({"token": "mytoken", "cached": False}),
        status_code=200,
        reason="OK",
    )

    q = Query(sql="SELECT * FROM mytable", ttl_minutes=5)  # type: ignore

    result = api.create_query(q)

    assert result.data is not None
    assert result.data.token == "mytoken"
    assert result.data.cached is False
    assert result.status_code == 200


def test_create_query_user_error(requests_mock):
    api = API("https://api.flipsidecrypto.xyz", "api_key")

    result = requests_mock.post(
        api.get_url("queries"),
        text=json.dumps({"errors": "user_error"}),
        status_code=400,
        reason="User Error",
    )

    q = Query(sql="SELECT * FROM mytable", ttl_minutes=5)  # type: ignore

    result = api.create_query(q)
    assert result.data is None
    assert result.status_msg == "User Error"
    assert result.status_code == 400
    assert result.error_msg == "user_error"


def test_create_query_server_error(requests_mock):
    api = API("https://api.flipsidecrypto.xyz", "api_key")

    result = requests_mock.post(api.get_url("queries"), status_code=500, reason="Server Error")

    q = Query(sql="SELECT * FROM mytable", ttl_minutes=5)  # type: ignore

    result = api.create_query(q)
    assert result.data is None
    assert result.status_msg == "Server Error"
    assert result.status_code == 500
    assert result.error_msg is None


def getQueryResultSetData(status: str) -> QueryResultJson:
    return QueryResultJson(
        queryId="test",
        status=status,
        results=[
            [1, "0x-tx-id-0", "0xfrom-address-0", True, 0.5],
            [2, "0x-tx-id-1", "0xfrom-address-1", False, 0.75],
            [3, "0x-tx-id-2", "0xfrom-address-2", False, 1.75],
            [4, "0x-tx-id-3", "0xfrom-address-3", True, 100.001],
        ],
        startedAt="2022-05-19T00:00:00.00Z",
        endedAt="2022-05-19T00:01:30.00Z",
        columnLabels=[
            "block_id",
            "tx_id",
            "from_address",
            "succeeded",
            "amount",
        ],
        columnTypes=["number", "string", "string", "boolean", "number"],
        message="",
        errors=None,
        pageSize=100,
        pageNumber=0,
    )


def test_get_query_result(requests_mock):
    api = API("https://api.flipsidecrypto.xyz", "api_key")
    query_id = "test_query_id"
    page_number = 1
    page_size = 10

    query_result_json = getQueryResultSetData(QueryStatus.Finished).dict()

    result = requests_mock.get(
        api.get_url(f"queries/{query_id}"),
        text=json.dumps(query_result_json),
        status_code=200,
        reason="OK",
    )

    result = api.get_query_result(query_id, page_number, page_size)
    assert result.data is not None
    assert result.status_code == 200


def test_get_query_result_user_error(requests_mock):
    api = API("https://api.flipsidecrypto.xyz", "api_key")
    query_id = "test_query_id"
    page_number = 1
    page_size = 10

    result = requests_mock.get(
        api.get_url(f"queries/{query_id}"),
        text=json.dumps({"errors": "user_error"}),
        status_code=400,
        reason="User Error",
    )

    result = api.get_query_result(query_id, page_number, page_size)
    assert result.data is None
    assert result.status_msg == "User Error"
    assert result.status_code == 400
    assert result.error_msg == "user_error"


def test_get_query_result_server_error(requests_mock):
    api = API("https://api.flipsidecrypto.xyz", "api_key")
    query_id = "test_query_id"
    page_number = 1
    page_size = 10

    result = requests_mock.get(api.get_url(f"queries/{query_id}"), status_code=500, reason="Server Error")

    result = api.get_query_result(query_id, page_number, page_size)
    assert result.data is None
    assert result.status_msg == "Server Error"
    assert result.status_code == 500
    assert result.error_msg is None



import json

from shroomdk.api import API
from shroomdk.errors import (
    QueryRunExecutionError,
    QueryRunTimeoutError,
    SDKError,
    ServerError,
    UserError,
)
from shroomdk.integrations.query_integration import QueryIntegration
from shroomdk.integrations.query_integration.query_integration import DEFAULTS
from shroomdk.models import Query, QueryStatus
from shroomdk.models.api import QueryResultJson


def get_api():
    return API("https://api.flipsidecrypto.xyz", "api_key")


def test_query_defaults():
    qi = QueryIntegration(get_api())

    # Test that the defaults are semi-overridden
    q = Query(sql="", ttl_minutes=5, page_number=5, page_size=10)  # type: ignore
    next_q = qi._set_query_defaults(q)

    assert next_q.page_number == 5
    assert next_q.page_size == 10
    assert next_q.ttl_minutes == 5
    assert next_q.cached == DEFAULTS.cached
    assert next_q.timeout_minutes == DEFAULTS.timeout_minutes

    # Test that the defaults are not overridden
    q = Query(sql="")  # type: ignore
    next_q = qi._set_query_defaults(q)

    assert next_q.page_number == DEFAULTS.page_number
    assert next_q.page_size == DEFAULTS.page_size
    assert next_q.ttl_minutes == DEFAULTS.ttl_minutes
    assert next_q.cached == DEFAULTS.cached
    assert next_q.timeout_minutes == DEFAULTS.timeout_minutes


def test_run_failed_to_create_query(requests_mock):
    api = get_api()
    qi = QueryIntegration(api)

    # Test 400 error
    q = Query(sql="", ttl_minutes=5, page_number=5, page_size=10)  # type: ignore
    requests_mock.post(
        api.get_url("queries"),
        text=json.dumps({"errors": "user_error"}),
        status_code=400,
        reason="User Error",
    )

    try:
        qi.run(q)
    except UserError as e:
        assert type(e) == UserError

    # Test 500 error
    requests_mock.post(
        api.get_url("queries"),
        text=json.dumps({"errors": "server_error"}),
        status_code=500,
        reason="Server Error",
    )

    try:
        qi.run(q)
    except ServerError as e:
        assert type(e) == ServerError

    # Unknown SDK Error
    requests_mock.post(
        api.get_url("queries"),
        text=json.dumps({"errors": "unknown_error"}),
        status_code=300,
        reason="Unknown Error",
    )

    try:
        qi.run(q)
    except SDKError as e:
        assert type(e) == SDKError

    # No query run data
    requests_mock.post(api.get_url("queries"), status_code=200, reason="OK")

    try:
        qi.run(q)
    except SDKError as e:
        assert type(e) == SDKError


def test_get_query_result_server_errors(requests_mock):
    api = get_api()
    qi = QueryIntegration(api)

    api = API("https://api.flipsidecrypto.xyz", "api_key")
    query_id = "test_query_id"

    # User Error
    requests_mock.get(api.get_url(f"queries/{query_id}"), status_code=400, reason="user_error")

    try:
        qi._get_query_results("test_query_id")
    except UserError as e:
        assert type(e) == UserError

    # Server Error
    requests_mock.get(api.get_url(f"queries/{query_id}"), status_code=500, reason="server error")

    try:
        qi._get_query_results("test_query_id")
    except ServerError as e:
        assert type(e) == ServerError

    # SDK Error
    requests_mock.get(api.get_url(f"queries/{query_id}"), status_code=200, reason="ok")

    try:
        qi._get_query_results("test_query_id")
    except SDKError as e:
        assert type(e) == SDKError


def test_get_query_result_query_errors(requests_mock):
    api = get_api()
    qi = QueryIntegration(api)

    api = API("https://api.flipsidecrypto.xyz", "api_key")
    query_id = "test_query_id"
    page_number = 1
    page_size = 10

    # Query Status: Error
    query_result_json = getQueryResultSetData(QueryStatus.Error).dict()

    result = requests_mock.get(
        api.get_url(f"queries/{query_id}"),
        text=json.dumps(query_result_json),
        status_code=200,
        reason="OK",
    )

    try:
        result = qi._get_query_results(
            "test_query_id",
            page_number=page_number,
            page_size=page_size,
            attempts=0,
            timeout_minutes=1,
            retry_interval_seconds=0.0001,
        )
    except QueryRunExecutionError as e:
        assert type(e) == QueryRunExecutionError

    # Query Status: Finished
    query_result_json = getQueryResultSetData(QueryStatus.Finished).dict()

    result = requests_mock.get(
        api.get_url(f"queries/{query_id}"),
        text=json.dumps(query_result_json),
        status_code=200,
        reason="OK",
    )

    result = qi._get_query_results(
        "test_query_id",
        page_number=page_number,
        page_size=page_size,
        attempts=0,
        timeout_minutes=1,
        retry_interval_seconds=0.0001,
    )
    assert result.results is not None
    assert type(result.results) is list
    assert len(result.results) == len(query_result_json["results"])

    # Query Execution Error
    query_result_json = getQueryResultSetData(QueryStatus.Error).dict()

    result = requests_mock.get(
        api.get_url(f"queries/{query_id}"),
        text=json.dumps(query_result_json),
        status_code=200,
        reason="OK",
    )

    try:
        result = qi._get_query_results("test_query_id")
    except QueryRunExecutionError as e:
        assert type(e) == QueryRunExecutionError

    # Query Timeout
    query_result_json = getQueryResultSetData(QueryStatus.Pending).dict()

    result = requests_mock.get(
        api.get_url(f"queries/{query_id}"),
        text=json.dumps(query_result_json),
        status_code=200,
        reason="OK",
    )

    try:
        result = qi._get_query_results(
            "test_query_id",
            page_number=page_number,
            page_size=page_size,
            attempts=0,
            timeout_minutes=0.1,
            retry_interval_seconds=0.0001,
        )
    except QueryRunTimeoutError as e:
        assert type(e) == QueryRunTimeoutError


def getQueryResultSetData(status: str) -> QueryResultJson:
    return QueryResultJson(
        queryId="test",
        status=status,
        results=[
            [1, "0x-tx-id-0", "0xfrom-address-0", True, 0.5],
            [2, "0x-tx-id-1", "0xfrom-address-1", False, 0.75],
            [3, "0x-tx-id-2", "0xfrom-address-2", False, 1.75],
            [4, "0x-tx-id-3", "0xfrom-address-3", True, 100.001],
        ],
        startedAt="2022-05-19T00:00:00.00Z",
        endedAt="2022-05-19T00:01:30.00Z",
        columnLabels=[
            "block_id",
            "tx_id",
            "from_address",
            "succeeded",
            "amount",
        ],
        columnTypes=["number", "string", "string", "boolean", "number"],
        message="",
        errors=None,
        pageSize=100,
        pageNumber=0,
    )



from typing import Union

from shroomdk.api import API
from shroomdk.errors import (
    QueryRunExecutionError,
    QueryRunTimeoutError,
    SDKError,
    ServerError,
    UserError,
)
from shroomdk.models import (
    Query,
    QueryDefaults,
    QueryResultSet,
    QueryStatus,
    SleepConfig,
)
from shroomdk.models.api import QueryResultJson
from shroomdk.utils.sleep import get_elapsed_linear_seconds, linear_backoff

from .query_result_set_builder import QueryResultSetBuilder

DEFAULTS: QueryDefaults = QueryDefaults(
    ttl_minutes=60,
    cached=True,
    timeout_minutes=20,
    retry_interval_seconds=0.5,
    page_size=100000,
    page_number=1,
)


class QueryIntegration(object):
    def __init__(self, api: API, defaults: QueryDefaults = DEFAULTS):
        self.api = api
        self.defaults = defaults

    def run(self, query: Query) -> QueryResultSet:
        query = self._set_query_defaults(query)

        created_query = self.api.create_query(query)
        if created_query.status_code > 299:
            if created_query.status_code < 500 and created_query.status_code >= 400:
                raise UserError(created_query.status_code, created_query.error_msg)
            elif created_query.status_code >= 500:
                raise ServerError(created_query.status_code, created_query.error_msg)
            else:
                raise SDKError(
                    f"unknown SDK error when calling `api.create_query`, {created_query.error_msg}"
                )

        query_run = created_query.data
        if not query_run:
            raise SDKError("expected `created_query.data` from server but got `None`")

        query_results = self._get_query_results(
            query_run.token,
            page_number=query.page_number,
            page_size=query.page_size,
            timeout_minutes=query.timeout_minutes if query.timeout_minutes else 20,
            retry_interval_seconds=query.retry_interval_seconds
            if query.retry_interval_seconds
            else 1,
        )

        return QueryResultSetBuilder(query_results).build()

    def _set_query_defaults(self, query: Query) -> Query:
        query_default_dict = self.defaults.dict()
        query_dict = query.dict()
        query_default_dict.update(
            {k: v for (k, v) in query_dict.items() if v is not None}
        )
        return Query(**query_default_dict)

    def _get_query_results(
        self,
        query_run_id: str,
        page_number: int = 1,
        page_size: int = 100000,
        attempts: int = 0,
        timeout_minutes: Union[int, float] = 20,
        retry_interval_seconds: Union[int, float] = 1.0,
    ) -> QueryResultJson:

        query_run = self.api.get_query_result(query_run_id, page_number, page_size)
        status_code = query_run.status_code

        if status_code > 299:
            error_msg = query_run.status_msg if query_run.status_msg else "error"
            if query_run.error_msg:
                error_msg = query_run.error_msg
            if status_code >= 400 and status_code <= 499:
                raise UserError(status_code, error_msg)
            elif status_code >= 500:
                raise ServerError(status_code, error_msg)

        if not query_run.data:
            raise SDKError(
                "valid status msg returned from server but no data exists in the response"
            )

        query_status = query_run.data.status

        if query_status == QueryStatus.Finished:
            return query_run.data

        if query_status == QueryStatus.Error:
            raise QueryRunExecutionError()

        should_continue = linear_backoff(
            SleepConfig(
                attempts=attempts,
                timeout_minutes=timeout_minutes,
                interval_seconds=retry_interval_seconds,
            )
        )

        if not should_continue:
            elapsed_seconds = get_elapsed_linear_seconds(
                SleepConfig(
                    attempts=attempts,
                    timeout_minutes=timeout_minutes,
                    interval_seconds=retry_interval_seconds,
                )
            )

            raise QueryRunTimeoutError(elapsed_seconds)

        return self._get_query_results(
            query_run_id,
            page_number,
            page_size,
            attempts + 1,
            timeout_minutes,
            retry_interval_seconds,
        )



from shroomdk.models.query_status import (
    QueryStatus,
    QueryStatusError,
    QueryStatusFinished,
    QueryStatusPending,
)


def test_query_status():
    assert QueryStatusFinished == "finished"
    assert QueryStatusPending == "pending"
    assert QueryStatusError == "error"

    assert QueryStatus.Finished == "finished"
    assert QueryStatus.Pending == "pending"
    assert QueryStatus.Error == "error"



from .query_run_errors import (  # noqa: F401
    QueryRunExecutionError,
    QueryRunRateLimitError,
    QueryRunTimeoutError,
)
from .sdk_error import SDKError  # noqa: F401
from .server_error import ServerError  # noqa: F401
from .user_error import UserError  # noqa: F401



from typing import Optional, Union

from pydantic import BaseModel, Field


class Query(BaseModel):
    sql: str = Field(None, description="SQL query to execute")
    ttl_minutes: Optional[int] = Field(None, description="The number of minutes to cache the query results")
    timeout_minutes: Optional[int] = Field(None, description="The number of minutes to timeout the query")
    retry_interval_seconds: Optional[Union[int, float]] = Field(1, description="The number of seconds to use between retries")
    cached: Optional[bool] = Field(
        None,
        description="An override on the cache. A value of true will Re-Execute the query.",
    )
    page_size: int = Field(None, description="The number of results to return per page")
    page_number: int = Field(None, description="The page number to return")



import json
from typing import List

import requests
from requests.adapters import HTTPAdapter, Retry

from .models import Query
from .models.api import (
    CreateQueryJson,
    CreateQueryResp,
    QueryResultJson,
    QueryResultResp,
)


class API(object):
    def __init__(
        self,
        base_url: str,
        api_key: str,
        max_retries: int = 10,
        backoff_factor: float = 1,
        status_forcelist: List[int] = [429, 500, 502, 503, 504],
        method_allowlist: List[str] = [
            "HEAD",
            "GET",
            "PUT",
            "POST",
            "DELETE",
            "OPTIONS",
            "TRACE",
        ],
    ):
        self._base_url = base_url
        self._api_key = api_key

        # Session Settings
        self._MAX_RETRIES = max_retries
        self._BACKOFF_FACTOR = backoff_factor
        self._STATUS_FORCE_LIST = status_forcelist
        self._METHOD_ALLOWLIST = method_allowlist

    def get_url(self, path: str) -> str:
        return f"{self._base_url}/{path}"

    def create_query(self, query: Query) -> CreateQueryResp:
        result = self._session.post(
            self.get_url("queries"),
            data=json.dumps(query.dict()),
            headers=self._headers,
        )

        try:
            data = result.json()
        except json.decoder.JSONDecodeError:
            data = None

        return CreateQueryResp(
            status_code=result.status_code,
            status_msg=result.reason,
            error_msg=data.get("errors") if data else None,
            data=CreateQueryJson(**data)
            if data and data.get("errors") is None
            else None,
        )

    def get_query_result(
        self, query_id: str, page_number: int, page_size: int
    ) -> QueryResultResp:
        result = self._session.get(
            self.get_url(f"queries/{query_id}"),
            params={"pageNumber": page_number, "pageSize": page_size},
            headers=self._headers,
        )

        try:
            data = result.json()
        except json.decoder.JSONDecodeError:
            data = None

        return QueryResultResp(
            status_code=result.status_code,
            status_msg=result.reason,
            error_msg=data.get("errors") if data else None,
            data=QueryResultJson(**data)
            if data and data.get("errors") is None
            else None,
        )

    @property
    def _headers(self) -> dict:
        return {
            "Accept": "application/json",
            "Content-Type": "application/json",
            "x-api-key": self._api_key,
        }

    @property
    def _session(self) -> requests.Session:
        if hasattr(self, "__session"):
            return self._session

        retry_strategy = Retry(
            total=self._MAX_RETRIES,
            backoff_factor=self._BACKOFF_FACTOR,
            status_forcelist=self._STATUS_FORCE_LIST,
            allowed_methods=self._METHOD_ALLOWLIST,
        )

        adapter = HTTPAdapter(max_retries=retry_strategy)
        http = requests.Session()
        http.mount("https://", adapter)
        http.mount("http://", adapter)

        self.__session = http
        return self.__session



from datetime import datetime
from typing import List, Union

from shroomdk.models import QueryResultSet, QueryRunStats
from shroomdk.models.api import QueryResultJson


class QueryResultSetBuilder(object):
    def __init__(self, data: QueryResultJson):
        self.query_id = data.queryId
        self.status = data.status
        self.columns = data.columnLabels
        self.column_types = data.columnTypes
        self.rows = data.results

        self.run_stats = self.compute_run_stats(data)
        self.records = self.create_records(data)

    def build(self) -> QueryResultSet:
        return QueryResultSet(
            query_id=self.query_id,
            status=self.status,
            columns=self.columns,
            column_types=self.column_types,
            rows=self.rows,
            run_stats=self.run_stats,
            records=self.records,
            error=None,
        )

    def compute_run_stats(self, data: QueryResultJson) -> QueryRunStats:
        if not data.startedAt or not data.endedAt:
            raise Exception("Query has no data ")
        start_time = datetime.strptime(data.startedAt, "%Y-%m-%dT%H:%M:%S.%fZ")
        end_time = datetime.strptime(data.endedAt, "%Y-%m-%dT%H:%M:%S.%fZ")
        return QueryRunStats(
            started_at=start_time,
            ended_at=end_time,
            elapsed_seconds=(end_time - start_time).seconds,
            record_count=len(data.results) if data.results else 0,
        )

    def create_records(self, data: QueryResultJson) -> Union[List[dict], None]:
        if not data or not data.results:
            return None

        column_labels = data.columnLabels
        if not column_labels:
            return None

        records: List[dict] = []
        for row in data.results:
            if not row:
                continue

            record = {}
            for i, col in enumerate(column_labels):
                record[col.lower()] = row[i]

            records.append(record)

        return records



from .api_response import ApiResponse  # noqa: F401
from .create_query_resp import CreateQueryJson, CreateQueryResp  # noqa: F401
from .query_result_resp import QueryResultJson, QueryResultResp  # noqa: F401



from typing import Any, Union

from pydantic import BaseModel


class ApiResponse(BaseModel):
    status_code: int
    status_msg: Union[str, None]
    error_msg: Union[str, None]
    data: Union[Any, None]



from datetime import datetime

from pydantic import BaseModel, Field


class QueryRunStats(BaseModel):
    started_at: datetime = Field(None, description="The start time of the query run.")
    ended_at: datetime = Field(None, description="The end time of the query run.")
    elapsed_seconds: int = Field(
        None,
        description="The number of seconds elapsed between the start and end times.",
    )
    record_count: int = Field(None, description="The number of records returned by the query.")



from typing import Any, List, Optional, Union

from pydantic import BaseModel

from .api_response import ApiResponse


class QueryResultJson(BaseModel):
    queryId: Optional[str]
    status: str
    results: Optional[List[Any]]
    startedAt: Optional[str]
    endedAt: Optional[str]
    columnLabels: Optional[List[str]]
    columnTypes: Optional[List[str]]
    message: Optional[str]
    errors: Optional[str]
    pageNumber: Optional[int]
    pageSize: Optional[int]


class QueryResultResp(ApiResponse):
    data: Union[QueryResultJson, None]



from pydantic import BaseModel, Field


class QueryDefaults(BaseModel):
    ttl_minutes: int = Field(None, description="The number of minutes to cache the query results")
    cached: bool = Field(False, description="Whether or not to cache the query results")
    timeout_minutes: int = Field(None, description="The number of minutes to timeout the query")
    retry_interval_seconds: float = Field(None, description="The number of seconds to wait before retrying the query")
    page_size: int = Field(None, description="The number of results to return per page")
    page_number: int = Field(None, description="The page number to return")



from typing import Optional, Union

from pydantic import BaseModel, Field

from .api_response import ApiResponse


class CreateQueryJson(BaseModel):
    token: str = Field(None, description="The server-side token of the query being executed.")
    errors: Union[Optional[str], None] = Field(False, description="Error that occured when creating the query.")
    cached: Optional[bool] = Field(False, description="Whether the query is cached or not.")


class CreateQueryResp(ApiResponse):
    data: Union[CreateQueryJson, None]
